# Error Handling {#error_handling}

@tableofcontents

Errors can show up in almost any location on this library, for example:
- When an incorrect Message is received, containing invalid information (due to a faulty channel or operator error)
- When an unexpected bug occurs
- When radiation or a hardware issue leads to an unexpected software state

This library incorporates an error handling framework to allow proper identification and recovery from various errors,
without violating the performance and memory constraints of an embedded system.

As [exceptions are not used](docs/writing_code.md), this library uses a slightly different approach to testing and
reporting errors from what is standard in C++.

@note Errors do not have to be fatal or lead to a loss of the subsystem. Most errors should be recoverable with minimal impact
to the rest of the system.

## The ErrorHandler class

The processing of errors in the `ecss-services` library is managed by the @ref ErrorHandler class. The ErrorHandler
provides a number of convenience functions to report any errors that occur using a single line of code (demonstrated below).

In order to convey the most information using the minimal amount of data, all **error types** are encoded into different
enumerations inside the @ref ErrorHandler.

When deemed necessary, the developer can add additional **auxiliary data** to each error, to provide more observability
to the ground station if needed.

@todo Auxiliary data not implemented yet, see [#28](https://gitlab.com/acubesat/obc/ecss-services/-/issues/28)

### The ST[01] service

The `ST[01] service request verification` service is responsible for generating error notifications when an issue is
found in a **received Telemetry message**. It does not cover errors generated by internal functions.

For example, the following incidents **will** trigger an ST[01] error:
- Ground Station asks for Parameter #205, but Parameter #205 does not exist
- Ground Station asks to disable Temperature Sensor, but Temperature Sensor is already disabled
- Ground Station sends a corrupted message that makes no sense

The following incidents **will not** trigger an ST[01] error:
- A division by zero happens
- A loop that should run 5 times runs 101 times
- A temperature sensor experiences a failure

Any ST[01] failure report (i.e. a TM message generated by ST[01] to inform us about a problem that was caused by a TC
message) is tied to **the TM that caused it**. That means that every ST[01] report contains information about the
problematic TM, such as its packet type and counter. The report also contains the **error type** and any **auxiliary
data** available for this error.

To report an error tied to a Message, you can use @ref ErrorHandler::reportError.
```cpp
ErrorHandler::reportError(message, errorType);
```

You can also use convenience functions like @ref Message::assertType, @ref Message::assertTM, or @ref Message::assertTC.
These will check if your Message is of a specific type, and call `reportError` on their own if it is not.

### Internal errors

Errors that are not tied to a specific TC and are caused independently by a microcontroller are termed _internal
errors_. While they are still managed by the @ref ErrorHandler, they are not part of the standard.

To report an internal error, you can use @ref ErrorHandler::reportInternalError.
```cpp
ErrorHandler::reportInternalError(errorType);
```

You can also use the convenience @ref ASSERT_INTERNAL macro, which checks if a condition is true, and calls @ref
ErrorHandler::reportInternalError automatically if not:
```cpp
ASSERT_INTERNAL(5 > 3, ErrorHandler::MathematicsIsBroken);
```

## Function results
An important question that comes to mind when disabling extensions is how errors are propagated through the code. When a
function is executed, it must be clear when the result of the function is correct and when it is not. Additionally, any
parts of the code that depend on a successful result may need to avoid being executed when an error occurs.

While there are different available approaches to failure propagation (e.g.
[errno](https://www.tutorialspoint.com/cprogramming/c_error_handling.htm), error pointers as arguments etc.), the
`ecss-services` library indicates the error in the **return value** of each function, through the @ref Result class.

The @ref Result class can contain **either**:
- A successful result (e.g. an `int` or a `string` or a @ref Parameter), or
- An error, explaining why it failed to find a result

By calling the relevant functions of a @ref Result, the user can access its contents on success, or abort the operation
in case of failure.

### Example usage

The simplest usage of a `Result` might be as follows:
```cpp
Result<T,E> something = ...

if (something.error()) {
    // No! An error happened! We cannot continue!
    return;
}

// No errors, we can happily use the result
something->coolFunction();
```

### Error propagation
The @ref Result class offers many convenience functions to aid in common, yet complex situations.

For example, we may want a function that takes the average of many parameters. However, this function will fail if one
of those parameters does not exist.

```cpp
Result<float, Error> getAverage(ParameterArray parameterIds) {
    // Don't allow entering an empty array, by throwing the relevant error
    if (parameterIds.empty()) {
        // Error automatically gets converted to Result<float, Error>
        return ErrorHandler::EmptyInputError;
    }

    float sum = 0;

    for (auto parameterId : parameterIds) {
        // You should use `auto` here. We keep the full type for demonstration purposes
        Result<Parameter, Error> parameter = ParameterService::getParameterById(parameterId);

        if (parameter.error()) {
            // Result<anything, Error> automatically gets converted to Result<float, Error>
            return parameter;
        }

        sum += parameter.getValue();
    }

    // float automatically gets converted to Result<float, Error>
    return sum / parameterIds.count();
}
```

@todo Result class not implemented and not used. See [#134](https://gitlab.com/acubesat/obc/ecss-services/-/issues/134)
